'use strict';

// Imports
var fs = require('fs');
var url = require('url');
var spawn = require('child_process').spawn;

// Wrong way to do this but still, it works here...
function extend(o, w) {
  for (var name in w) {
    if (!(name in o))
      o[name] = w[name];
  }
  return o;
}

// Simple trim (used to clean headers)
function trim(str) {
  return str.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
}

// Parse headers until an empty line is found
function HeaderParser(onHeader, onBody) {
  this.onHeader = onHeader;
  this.onBody = onBody;

  this.header_mode = true;

  this.header_name = null;
  this.buffer = '';
}
HeaderParser.prototype = Object.create(Object.prototype, {
  // Header separator
  sep: {
    enumerable: true,
    value: ':'.charCodeAt()
  },
  // EOL char1
  el1: {
    enumerable: true,
    value: '\r'.charCodeAt()
  },
  // EOL char2
  el2: {
    enumerable: true,
    value: '\n'.charCodeAt()
  },

  // Parse some data from a buffer and make the best of it
  _parse: {
    value: function(data) {
      var i = 0;

      var handleHeader = function(end) {
        // Jump over the next char
        ++i;

        // Advertise we found a header
        if (this.header_name)
          this.onHeader(this.header_name, trim(this.buffer));

        // If there is no header_name, means that we found and empty line
        // => turn off header parsing
        if (this.header_name === null) {
          this._headerModeOff(data, end);
          return false;
        }

        // Reset
        this.header_name = null;
        this.buffer = '';

        return true;
      }.bind(this);

      // If we already found data & last byte was EL1 and now we found as first byte EL2, we found a header end!
      if (this.buffer.length && data.length &&
          this.buffer[this.buffer.length - 1] == this.el1 &&
          data[0] == this.el2)
        // Handle header and stop if we reached end
        if (!handleHeader(i + 1)) return;

      // For each byte into the buffer
      for (; i < data.length; ++i) {
        // If we are parsing a header_name and found a separator
        if (this.header_name === null &&
            data[i] == this.sep) {
          // Set the header name
          this.header_name = trim(this.buffer);
          // Reset the buffer
          this.buffer = '';
        }
        // If we are parsing a header value and found a correct EOL
        else if (data[i] == this.el1 &&
                 i + 1 < data.length && data[i + 1] == this.el2) {
          // Handle header and stop if we reached end
          if (!handleHeader(i + 2))
            return;
        }
        // Otherwise append to buffer
        else
          this.buffer += String.fromCharCode(data[i]);
      }
    }
  },

  // Switch off header mode
  _headerModeOff: {
    value: function(data, end_index) {
      // Switch mode
      this.header_mode = false;

      // Get the leftover data
      var leftOver = new Buffer(data.length - end_index);
      data.copy(leftOver, 0, end_index);

      this.onBody(leftOver);
    }
  },

  // Get the on data function
  onData: {
    enumerable: true,
    value: function() {
      return function(data) {
        if (this.header_mode)
          return this._parse(data);
        this.onBody(data);
      }.bind(this);
    }
  }
});

// Generate the app for HTTP
function createApp(cgiPath, options) {
  options = options || {};

  // Set debug level
  var debug_lvl = 'debug' in options ? options.debug : 0;

  // Get cwd and insure it is an absolute path
  var cwd = process.cwd();
  if ('cwd' in options)
    cwd = (options.cwd[0] != '/' ? cwd + '/': '') + options.cwd;
  if (debug_lvl > 0)
    console.log('info  - mod_cgi - cgi cwd set to `' + cwd + '\'');

  // Insure cgiPath is absolute, use cwd to complete it if not
  if (cgiPath[0] != '/')
    cgiPath = cwd + '/' + cgiPath;
  // Check that the CGI exists and send an error if not
  if (!fs.existsSync(cgiPath))
    throw "error - mod_cgi - error: cgi file `" + cgiPath + "' not found.";
  if (debug_lvl > 0)
    console.log('info  - mod_cgi - cgi path set to `' + cgiPath + '\'');

  // Insure we have a scriptName option
  if (!('scriptName' in options))
    options.scriptName = '/';
  if (debug_lvl > 0)
    console.log('info  - mod_cgi - script name set to `' + options.scriptName + '\'');

  // Return the app function
  return function exec(req, res) {
    // Some default values
    var SERVER_SOFTWARE = "Node/" + process.version;
    var SERVER_PROTOCOL = "HTTP/1.1";
    var GATEWAY_INTERFACE = "CGI/1.1";

    // If there is no uri, parse it ourselves
    req.uri = req.uri || url.parse(req.url);

    // Parse host header to get precise info on host & port
    var host = req.headers.host.split(':');
    var hostname = host[0];
    var port = host[1] || 80;

    // Extend our env variable from global env
    var env = extend(extend({}, process.env));

    // Get the path to give the script
    var path = req.uri.pathname;
    // If there is a scriptName, remove it from the path
    if (options.scriptName.length > 1) {
      path = path.replace(options.scriptName, '');
      if (path[0] != '/') path = '/' + path;
    }

    // Add necessary CGI values
    env = extend(env, {
      GATEWAY_INTERFACE:  GATEWAY_INTERFACE,
      SCRIPT_NAME:        options.scriptName,
      PATH_INFO:          path,
      SERVER_NAME:        hostname,
      SERVER_PORT:        port,
      SERVER_PROTOCOL:    SERVER_PROTOCOL,
      SERVER_SOFTWARE:    SERVER_SOFTWARE,
      REQUEST_METHOD:     req.method,
      QUERY_STRING:       req.uri.query
    });

    // Complete those with some optional ones
    if ('content-length' in req.headers) {
      env.CONTENT_LENGTH = req.headers['content-length'];
    }
    if ('content-type' in req.headers) {
      env.CONTENT_TYPE = req.headers['content-type'];
    }
    if ('authorization' in req.headers) {
      var auth = req.headers.authorization.split(' ');
      env.AUTH_TYPE = auth[0];
    }

    // Transform all headers into a general looking env
    for (var header in req.headers) {
      // Env looks like HTTP_HEADER_ALL_CAPS
      var name = 'HTTP_' + header.toUpperCase().replace(/[^A-Z0-9_]/g, '_');
      env[name] = req.headers[header];
    }

    // Write env if in very verbose
    if (debug_lvl > 2) {
      console.log('debug - modcgi - env is:');
      console.log('==');
      for (var h in env)
        console.log('  ' + h + ': ' + env[h]);
      console.log('==');
    }

    // Write what we are loading if asked
    if (debug_lvl > 0)
      console.log('info  - mod_cgi - ' + env.REQUEST_METHOD +'|' +
                  env.PATH_INFO + '|' + env.QUERY_STRING);


    // Run CGI
    var cgi = spawn(cgiPath, [], {
      cwd: options.cwd,
      env: env
    });
    req.pipe(cgi.stdin);

    // Parse headers
    var hp = new HeaderParser(
      // On header received
      function(name, val) {
        // Write headers on console if debug level high enough
        if (debug_lvl > 1)
          console.log('debug - mod_cgi - Header|' + name + '|' + val);

        // If status, set statusCode
        if (name == 'Status')
          res.statusCode = parseInt(val);
        // Otherwise, set the header
        else
          res.setHeader(name, val);
      },
      // After headers, send everything into res.write
      res.write.bind(res)
    );

    // Plug the parser to stdout
    cgi.stdout.on('data', hp.onData());

    // On exit, detect errors to get the user some info
    cgi.on('exit', function(code) {
      var message = '';
      if (code) {
        console.log('error - mod_cgi - Script exited with value ' + code);
        res.writeHead(500);
        message = '<h1>Error during script execution. Abort.</h1>';
      }
      res.end(message);
    });
  }
}

// Export the createApp handler
module.exports.createApp = createApp;
